"""
Manages background tasks.
"""
from asyncio import Task, CancelledError, Queue, create_task, gather
from collections.abc import Coroutine
from typing import Any, Callable, Optional
from aiohttp import web
import logging

class BackgroundTasks:
    """
    Manages asyncio background tasks. An instance of this class is available as an application property with name
    HEA_BACKGROUND_TASKS. BackgroundTasks objects are truthy if having managed any tasks and falsy otherwise.
    """

    def __init__(self, app: web.Application | None = None, queue_size= 0):
        """
        Creates a background task manager with the provided queue size.

        :param queue_size: the queue size. A value of zero means infinite size.
        """
        self.__app = app
        self.__tasks: dict[str, Task] = dict()
        self.__done: dict[str, Exception | None] = dict()
        self.__queue = Queue[Task](queue_size)
        self.__metadata: dict[str, dict[str, Any]] = dict()
        self.__result: dict[str, Any] = dict()
        self.__cancelled: dict[str, bool] = dict()
        self.__logger = logging.getLogger(__name__)

    def contains(self, name: str) -> bool:
        """
        Whether this background task manager is managing a task with the given name.

        :param name: the name of the background task.
        :return True or False.
        """
        return name in self.__tasks or name in self.__done

    async def add(self, coro: Callable[[Optional[web.Application]], Coroutine[Any, Any, None]], name: str | None = None, **metadata) -> str:
        """
        A coroutine for creating a new asyncio task from the given coroutine and adds it to the tasks managed by this
        task manager. If a name is provided, the task will be given the provided name. Otherwise, a name will be
        generated automatically.

        :param coro: the coroutine (required).
        :param name: an optional name.
        :param metadata: any key-value pairs associated with the task.
        :return: the name of the created task.
        """
        self.__logger.debug('Adding coroutine %s...', name)
        task = create_task(coro(self.__app), name=name)
        name_ = task.get_name()
        self.__tasks[name_] = task
        self.__metadata[name_] = metadata
        await self.__queue.put(task)
        self.__logger.debug('Coroutine %s added.', name_)
        return name_

    def metadata(self, name: str, key: str) -> Any:
        """
        Retrieve task metadata.

        :param name: the name of the task.
        :param key: the key of the metadata value of interest.
        :return: the metadata value.
        :raises KeyError: if the given task has no metadata with the given key.
        """
        return self.__metadata[name][key]

    def cancel(self, name: str):
        """
        Requests cancellation of the task with the given name.

        :param name: the name of the task to cancel.
        """
        self.__tasks[name].cancel()

    def cancel_all(self):
        """
        Requests cancellation of all tasks managed by this background task manager.
        """
        for task in self.__tasks.values():
            task.cancel()

    async def join(self, name: str):
        """
        A coroutine for awaiting completion of the task with the given name. This can be called repeatedly for the same
        task; if the task is already done, this coroutine does nothing. An exception raised by a task is logged as an
        error and not re-raised. CancelledError is re-raised after marking the task as done with no result. Attempting
        to join a non-existent task does nothing.

        :param name: the name of the task.
        :raises KeyError: if no task with the given name is managed by this background task manager.
        """
        try:
            await self.__tasks[name]
        except CancelledError:
            raise
        except:  # self._mark_done will check the task's result and handle any exceptions.
            pass
        finally:
            self._mark_done(name)

    async def join_all(self) -> None:
        """
        A coroutine for awaiting completion of all in-progress tasks managed by this backround task manager. Exceptions
        raised by tasks are logged as errors and not re-raised.
        """
        tasks = tuple((k, v) for k, v in self.__tasks.items())
        try:
            await gather(*(task for _, task in tasks))
        except CancelledError:
            raise
        except:  # self._mark_done will check each task's result and handle any exceptions.
            pass
        finally:
            self.cancel_all()
            await gather(*(task for _, task in tasks), return_exceptions=True)
            for name, _ in tasks:
                self._mark_done(name)

    def in_progress(self, name: str) -> bool:
        """
        Returns whether the task is managed by this task manager and is still running.

        :param name: the name of the task.
        :return: True or False.
        """
        return self.contains(name) and not self.done(name)

    def done(self, name: str) -> bool:
        """
        Returns whether the task with the given name is completed.

        :param name: the name of the task to check.
        :return: True or False.
        """
        return name in self.__done

    def succeeded(self, name: str) -> bool:
        """
        Returns whether the task with the given name completed successfully.

        :param name: the name of the task to check.
        :return: True or False.
        """
        return bool(self.done(name) and not self.error(name) and not self.cancelled(name))

    def failed(self, name: str) -> bool:
        """
        Returns whether the task with the given name failed.

        :param name: the name of the task to check.
        :return: True or False.
        """
        return bool(self.done(name) and self.error(name))

    def error(self, name: str) -> Exception | None:
        """
        Returns the exception raised within a failed task.

        :param name: the name of the task to check.
        :return: An exception if the task failed, or None if it succeeded.
        :raises ValueError: if the task is not done or not managed by this background task manager.
        """
        if name not in self.__done:
            raise ValueError(f'{name} not done or not managed by this background task manager.')
        return self.__done[name]

    def cancelled(self, name: str) -> bool:
        """
        Returns whether the task with the given name was cancelled.

        :param name: the name of the task to check.
        :return: True if the task was cancelled, False otherwise.
        :raises ValueError: if the task is not done or not managed by this background task manager.
        """
        if name not in self.__done:
            raise ValueError(f'{name} not done or not managed by this background task manager.')
        return self.__cancelled.get(name, False)

    def result(self, name: str) -> Any:
        if name not in self.__done:
            raise ValueError(f'{name} not done or not managed by this background task manager.')
        return self.__result[name]

    def remove(self, name: str):
        """
        Removes the task with the given name from this task manager. This method does nothing if no task with the given
        name is managed by this background task manager.

        :param name: the name of the task to remove.
        """
        if name in self.__tasks:
            del self.__tasks[name]
        if name in self.__done:
            del self.__done[name]
        if name in self.__metadata:
            del self.__metadata[name]
        if name in self.__cancelled:
            del self.__cancelled[name]
        if name in self.__result:
            del self.__result[name]

    async def clear(self):
        """
        A coroutine that cancels all running tasks and cleans up all resources.
        """
        self.cancel_all()
        await self.join_all()
        while not self.__queue.empty():
            self.__queue.get_nowait()
            self.__queue.task_done()
        self.__tasks.clear()
        self.__done.clear()
        self.__metadata.clear()
        self.__cancelled.clear()
        self.__result.clear()

    async def auto_join(self):
        """
        A coroutine that automatically executes tasks added to this task manager. It runs forever. When wrapped in a
        task, cancelling the task attempts to end all the managed tasks and ends this coroutine.
        """
        async def consumer():
            while task := await self.__queue.get():
                await self.join(task.get_name())
                self.__queue.task_done()


        consumers = [create_task(consumer()) for _ in range(5)]
        try:
            await gather(*consumers)
        except CancelledError:
            for c in consumers:
                c.cancel()
            await gather(*consumers, return_exceptions=True)
            raise



    def __len__(self) -> int:
        """
        Returns the number of tasks managed by this background task manager.

        :return: the number of tasks managed by this background task manager.
        """
        return len(self.__tasks) + len(self.__done)

    def running_tasks(self) -> int:
        """
        Returns the number of running tasks.

        :return: the number of running tasks.
        """
        return len(self.__tasks)

    def done_tasks(self) -> int:
        """
        Returns the number of done tasks.

        :return: the number of done tasks.
        """
        return len(self.__done)

    def _mark_done(self, name: str):
        """
        Marks the task with the given name as done, and logs any exception raised within the task. If there is no task
        with the given name, this method does nothing. If the task was cancelled, it is marked as done with no result,
        and the CancelledError is NOT re-raised. If the task raised an exception, it is marked as done with the
        exception as the result, and the exception is logged as an error. If the task completed successfully, it is
        marked as done with its result.

        :param name: the name of the task to mark done.
        """
        try:
            the_task = self.__tasks.get(name)
            if the_task is None:
                return
            self.__result[name] = the_task.result()
            self.__done[name] = None
            self.__cancelled[name] = False
            self.__logger.debug('Task %s completed successfully', name)
        except CancelledError:
            self.__done[name] = None
            self.__result[name] = None
            self.__cancelled[name] = True
        except Exception as e:
            self.__done[name] = e
            self.__cancelled[name] = False
            self.__result[name] = None
            self.__logger.exception('Task %s failed with exception', name)
        finally:
            self.__tasks.pop(name, None)

    def __str__(self) -> str:
        return f'Background task manager: in-progress tasks: {", ".join(self.__tasks)}; done tasks: {", ".join(self.__done)}'



"""
Tests for the SensitiveDataFilter logging filter. It tests with a logger named as this module's __name__.
"""
from abc import ABC, abstractmethod
import logging
from unittest import TestCase

from yarl import URL
from heaserver.service.logging import ScrubbingLogger, SensitiveDataFilter
from heaobject.keychain import AWSCredentials


log_record = {
    'user': 'alice',
    'password': 'mysecretpassword',
    'token': 'abcd1234',
    'info': 'This is a test log'
}

log_record_list = [log_record]

EXPECTED = "User login attempt: {'user': 'alice', 'password': '******', 'token': '******', 'info': 'This is a test log'}\n"


class AbstractTestLogging(TestCase, ABC):

    @abstractmethod
    def __init__(self, filter: logging.Filter, methodName='runTest'):
        super().__init__(methodName=methodName)
        self.filter = filter

    def setUp(self):
        self.logger = ScrubbingLogger(__name__)
        self.logger.propagate = False
        self.logger.setLevel(logging.INFO)
        self.stream_handler = logging.StreamHandler()
        self.stream_handler.addFilter(self.filter)
        self.logger.addHandler(self.stream_handler)

    def tearDown(self):
        try:
            self.stream_handler.close()
        finally:
            self.stream_handler.removeFilter(self.filter)
            self.logger.removeHandler(self.stream_handler)
            # There's no way to remove a logger from the logging module's internal registry,

    def _read_output_keys(self) -> str:
        return self._read_output('User login attempt: %r', log_record)

    def _read_output_str(self) -> str:
        return self._read_output(f'User login attempt: {log_record}')

    def _read_output_keys_list(self) -> str:
        return self._read_output('User login attempt: %r', log_record_list)

    def _read_output(self, str, args=None) -> str:
        self.logger.info(str, args) if args else self.logger.info(str)
        stream = self.stream_handler.stream
        stream.seek(0)
        return stream.read()


class TestKeyScrubbing(AbstractTestLogging):
    def __init__(self, methodName='runTest'):
        super().__init__(SensitiveDataFilter(), methodName=methodName)

    def test_logging_sensitive_keys(self):
        self.assertEqual(EXPECTED, self._read_output_keys())

class TestStrScrubbing(AbstractTestLogging):
    def __init__(self, methodName='runTest'):
        super().__init__(SensitiveDataFilter(), methodName=methodName)

    def test_logging_sensitive_str(self):
        self.assertEqual(EXPECTED, self._read_output_str())

    def test_logging_sensitive_str_regex(self):
        message = "Getting content at %s with headers"
        url = 'http://heaserver-volumes:8080/volumes/byfilesystemtype/heaobject.volume.AWSFileSystem?account_id=heaobject.account.AWSAccount%5E123456678901'
        expected = "Getting content at http://heaserver-volumes:8080/volumes/byfilesystemtype/heaobject.volume.AWSFileSystem?account_id=****** with headers\n"
        self.assertEqual(expected, self._read_output(message, url))

    def test_logging_sensitive_URL_regex(self):
        message = "Getting content at %s with headers"
        url = URL('http://heaserver-volumes:8080/volumes/byfilesystemtype/heaobject.volume.AWSFileSystem?account_id=heaobject.account.AWSAccount%5E123456678901')
        expected = "Getting content at http://heaserver-volumes:8080/volumes/byfilesystemtype/heaobject.volume.AWSFileSystem?account_id=****** with headers\n"
        self.assertEqual(expected, self._read_output(message, url))


class TestListScrubbing(AbstractTestLogging):
    def __init__(self, methodName='runTest'):
        super().__init__(SensitiveDataFilter(), methodName=methodName)

    def test_logging_sensitive_list(self):
        expected = "User login attempt: [{'user': 'alice', 'password': '******', 'token': '******', 'info': 'This is a test log'}]\n"
        self.assertEqual(expected, self._read_output_keys_list())


class CustomSensitiveDataFilter(SensitiveDataFilter):
    sensitive_keys = SensitiveDataFilter.sensitive_keys | {'user'}
EXPECTED_CUSTOM = "User login attempt: {'user': '******', 'password': '******', 'token': '******', 'info': 'This is a test log'}\n"

class TestCustomKeyList(AbstractTestLogging):
    def __init__(self, methodName='runTest'):
        super().__init__(CustomSensitiveDataFilter(), methodName=methodName)

    def test_logging_sensitive_keys_extra_keys(self):
        self.assertEqual(EXPECTED_CUSTOM, self._read_output_keys())

class TestInStrCustomKeyList(AbstractTestLogging):
    def __init__(self, methodName='runTest'):
        super().__init__(CustomSensitiveDataFilter(), methodName=methodName)

    def test_logging_sensitive_str_extra_keys(self):
        self.assertEqual(EXPECTED_CUSTOM, self._read_output_str())


class AWSCredentialsLogging(AbstractTestLogging):
    def __init__(self, methodName='runTest'):
        super().__init__(SensitiveDataFilter(), methodName=methodName)

    def test_logging_aws_credentials(self):
        self.maxDiff = None
        c: AWSCredentials = AWSCredentials()
        c.password = 'mysecretpassword'
        c.session_token = 'abcd1234'
        actual = self._read_output('User login attempt: %r', c)
        self.assertIn("'display_name': '******'", actual)
        self.assertIn("'session_token': '******'", actual)
        self.assertIn("'password': '******'", actual)
        self.assertIn("'managed': False", actual)

class SequenceOfAWSCredentialsLogging(AbstractTestLogging):
    def __init__(self, methodName='runTest'):
        super().__init__(SensitiveDataFilter(), methodName=methodName)
        self.maxDiff = None
        self.c1: AWSCredentials = AWSCredentials()
        self.c1.password = 'mysecretpassword1'
        self.c1.session_token = 'abcd1234'
        self.c2: AWSCredentials = AWSCredentials()
        self.c2.password = 'mysecretpassword2'
        self.c2.session_token = 'efgh5678'

    def test_logging_list_of_aws_credentials(self):
        actual = self._read_output('User login attempt: %r', [self.c1, self.c2])
        self.assertIn("'display_name': '******'", actual)
        self.assertIn("'session_token': '******'", actual)
        self.assertIn("'password': '******'", actual)
        self.assertIn("'managed': False", actual)

    def test_logging_tuple_of_aws_credentials(self):
        actual = self._read_output('User login attempt: %r', (self.c1, self.c2))
        self.assertIn("'display_name': '******'", actual)
        self.assertIn("'session_token': '******'", actual)
        self.assertIn("'password': '******'", actual)
        self.assertIn("'managed': False", actual)

class TestNonStrings(AbstractTestLogging):
    def __init__(self, methodName='runTest'):
        super().__init__(SensitiveDataFilter(), methodName=methodName)

    def test_logging_integer(self):
        self.assertEqual('User login attempt: 4\n', self._read_output('User login attempt: %d', 4))

    def test_logging_decimal(self):
        self.assertEqual('User login attempt: 4.20\n', self._read_output('User login attempt: %.2f', 4.2))

class TestScrubRegexes(AbstractTestLogging):
    def __init__(self, methodName='runTest'):
        super().__init__(SensitiveDataFilter(), methodName=methodName)

    def test_logging_sensitive_regex(self):
        log_record = 'Received access_token=abcd1234 from user'
        expected = 'Received access_token=****** from user\n'
        self.assertEqual(expected, self._read_output(log_record))

    def test_logging_unsensitive_regex(self):
        log_record = 'No sensitive data here'
        expected = 'No sensitive data here\n'
        self.assertEqual(expected, self._read_output(log_record))



import configparser

from yarl import URL

from .defaults import DEFAULT_PORT, DEFAULT_BASE_URL
from .crypt import EncryptionDecryptionKeyGetter, SecretDecryption
from heaobject.encryption import Encryption
import os


class Configuration(EncryptionDecryptionKeyGetter):
    """
    Configuration information for the service.
    """
    def __init__(self,
                 base_url: str | URL = DEFAULT_BASE_URL,
                 port: int | str = DEFAULT_PORT,
                 config_file: str | None = None,
                 config_str: str | None = None):
        """
        Initializes the configuration object.

        :param base_url: the base URL of the service. Required.
        :param port: the port the service will listen on. Required.
        :param config_file: an optional INI file with configuration data.
        :param config_str: an optional configuration INI file as a string. Parsed after any config_file.
        """
        self.__base_url = URL(base_url) if base_url else DEFAULT_BASE_URL
        self.__port = int(port) if port else DEFAULT_PORT
        self.__parsed_config = configparser.ConfigParser()
        self.__config_file = config_file
        self.__config_str = config_str
        if config_file:
            self.__parsed_config.read(config_file)
        if config_str:
            self.__parsed_config.read_string(config_str)
        self.__parsed_config.read_dict({})

    @property
    def port(self) -> int:
        """
        The port this service will listen on.
        :return: a port number.
        """
        return self.__port

    @property
    def base_url(self) -> URL:
        """
        This service's base URL.
        :return: a URL.
        """
        return self.__base_url

    @property
    def parsed_config(self) -> configparser.ConfigParser:
        """
        Any configuration information parsed from an INI file or INI string.
        :return: a configparser.ConfigParser object.
        """
        result = configparser.ConfigParser()
        result.read_dict(self.__parsed_config)
        return result

    @property
    def config_file(self) -> str | None:
        """
        The path to a config file.
        :return: the config file path string, or None.
        """
        return self.__config_file

    @property
    def config_str(self) -> str | None:
        """
        A string containing an INI file, if provided in the constructor.
        :return: a config file string, or None.
        """
        return self.__config_str

    @property
    def default_section(self) -> configparser.SectionProxy | None:
        """
        The default section of the parsed configuration.

        :return: a configparser.SectionProxy object.
        """
        return self.__parsed_config['DEFAULT'] if 'DEFAULT' in self.__parsed_config else None

    @property
    def registry_url(self) -> str | None:
        """
        The registry URL, if specified in the configuration.

        :return: the registry URL string, or None.
        """
        return self.__get_default_section_property('Registry')

    @property
    def encryption_key_file(self) -> str | None:
        """
        The path to the encryption key file, if specified in the configuration. A ~ is expanded to the user's home
        directory for Windows compatibility.

        :return: the encryption key file path string, or None.
        """
        prop = self.__get_default_section_property('EncryptionKeyFile')
        return os.path.expanduser(prop) if prop else None

    @property
    def encryption_key_env_var(self) -> str | None:
        """
        The encryption key, if specified in the configuration as a property
        value. The encryption key should NOT be stored in the configuration
        file in production settings!

        :return: the encryption key as a str, or None.
        """
        return self.__get_default_section_property('EncryptionKey')

    def get_attribute_encryption_key(self) -> bytes | None:
        """
        Gets the encryption key from the encryption key file, if specified,
        or from the environment variable, if specified. If stored as an
        environment variable, it is returned as UTF-8 encoded bytes.

        :return: the encryption key as bytes, or None.
        """
        key_file = self.encryption_key_file
        if key_file and os.path.exists(key_file):
            with open(key_file, 'rb') as f:
                return f.read().strip()
        else:
            env_var = self.encryption_key_env_var
            if env_var:
                return env_var.encode('utf-8')
        return None

    def get_attribute_encryption(self) -> Encryption | None:
        """
        Gets the attribute encryption instance if an encryption key is available.

        :return: an Encryption instance, or None.
        """
        key = self.get_attribute_encryption_key()
        return Encryption(key) if key else None

    def get_secret_decryption_key(self) -> bytes | None:
        """
        Gets the decryption key. In this implementation, it is the same as the encryption key.

        :return: the decryption key as bytes, or None.
        """
        return self.get_attribute_encryption_key()

    def get_secret_decryption(self) -> SecretDecryption | None:
        """
        Gets a SecretDecryption instance if an encryption key is available.

        :return: a SecretDecryption instance, or None.
        """
        encryption_key = self.get_attribute_encryption_key()
        if encryption_key:
            return SecretDecryption(encryption_key)
        return None

    def __get_default_section_property(self, property_name: str) -> str | None:
        """
        Gets a property from the default section of the configuration.

        :param property_name: the property name.
        :return: the property value as a string, or None.
        """
        default_section = self.default_section
        return default_section.get(property_name) if default_section else None



"""
Properties that are loaded into the aiohttp request context.

HEA_WSTL_BUILDER - A heaserver.service.wstl.RuntimeWeSTLDocumentBuilder instance.
HEA_ATTRIBUTE_ENCRYPTION_KEY - The symmetric key used to encrypt and decrypt sensitive HEAObject attributes. It is
    populated automatically. It is also used at present to decrypt configuration properties, but this could change in
    the future.
"""

HEA_WSTL_BUILDER = 'HEA_wstl_document'
HEA_ATTRIBUTE_ENCRYPTION_KEY = 'HEA_attribute_encryption_key'



from collections.abc import Mapping
import logging
import re
from typing import Any, Sequence, overload
from heaobject.root import HEAObject, get_all_heaobject_subclasses
from heaobject.scrubber import scrubbed
from inspect import isabstract

from yarl import URL
from heaserver.service.oidcclaimhdrs import ACCESS_TOKEN, EMAIL, NAME, FAMILY_NAME, GIVEN_NAME, SESSION_STATE, \
    PREFERRED_USERNAME


# A tuple of regexes that will be replaced with '******' or (regex, replacement string) tuples.
_regex_patterns: tuple[str | tuple[str, str], ...] = ((r'(access_token=)[^&\s]+', r'\1******'),  # access_token query parameter.
                                                      (r'(account_id=)[^&\s]+', r'\1******'),  # account_id query parameter.
                                                      (r'(heaobject\.account\.AWSAccount\^)[^\']+', r'\1******'),  # scrub AWS account DesktopObject ids.
                                                      (r'(heaobject\.folder\.AWSS3BucketItem\^)[^\']+', r'\1******'),  # scrub AWS bucket names.
                                                      (r"^arn:(?P<Partition>[^:\n]*):(?P<Service>[^:\n]*):(?P<Region>[^:\n]*):(?P<AccountID>[^:\n]*):(?P<ResourceType>[^:/\n]*[:/])?(?P<Resource>.*)$", r"arn:\1:\2:\3:******:\5******"))

_regex_compiled_patterns = tuple((re.compile(pattern_sub[0]), pattern_sub[1]) if isinstance(pattern_sub, tuple)
                                 else re.compile(pattern_sub) for pattern_sub in _regex_patterns)

# Put anything here that should be considered sensitive keys but is not marked as such in HEAObject attribute metadata
# or might be a dictionary key.
_sensitive_keys_set = set(['headers', 'credentials', 'Authorization', 'token', 'Arn', 'Resource', 'Path', 'UserId',
                           'Account', 'bucket_id', 'path', 's3_uri', 'resource_type_and_id', 'firstName', 'username',
                           'lastName', 'userId', 'userName', 'access_token']) | \
    {ACCESS_TOKEN, EMAIL, NAME, FAMILY_NAME, GIVEN_NAME, SESSION_STATE, PREFERRED_USERNAME}


for cls in get_all_heaobject_subclasses(lambda x: not isabstract(x)):
    instance: HEAObject = cls()
    for attr_name, attr_metadata in instance.get_all_attribute_metadata().items():
        if attr_metadata.sensitive:
            _sensitive_keys_set.add(attr_name)


class SensitiveDataFilter(logging.Filter):
    """
    A logging filter that masks sensitive data in log messages. It looks for the keys defined in the sensitive_keys
    class variable in the record arguments (if they are a Mapping) and the log message itself.

    Subclass this filter and override the `patterns` and `sensitive_keys` attributes to customize the patterns and keys
    to be masked. The default regex patterns and sensitive keys are defined in the patterns and sensitive_keys
    constants.

    Sequence arguments are converted to lists, and Mapping arguments are converted to dicts.

    This class is based on https://dev.to/camillehe1992/mask-sensitive-data-using-python-built-in-logging-module-45fa.
    """
    patterns = _regex_compiled_patterns
    sensitive_keys = frozenset(_sensitive_keys_set)
    __sensitive_key_regexes = dict[str, tuple[re.Pattern[str], str]]()
    __ends_with_sensitive_keys = tuple('.' + k for k in sensitive_keys)

    def filter(self, record: logging.LogRecord) -> bool:
        try:
            record.args = self.__mask_sensitive_all_args(record.args)
            record.msg = self.__mask_sensitive_msg(record.msg)
            return True
        except:
            return True

    @overload
    def __mask_sensitive_all_args(self, args: None) -> None:
        pass

    @overload
    def __mask_sensitive_all_args(self, args: Mapping[str, object]) -> Mapping[str, object]:
        pass

    @overload
    def __mask_sensitive_all_args(self, args: tuple[object, ...]) -> tuple[object, ...]:
        pass

    def __mask_sensitive_all_args(self, args: tuple[object, ...] | Mapping[str, object] | None) -> tuple[object, ...] | Mapping[str, object] | None:
        """
        Masks sensitive data in all of the given log record arguments.

        :param args: the log record arguments, which can be a tuple or list.
        :return: the masked log record arguments.
        """
        if not isinstance(args, tuple):
            return self.__mask_sensitive_args(args)
        else:
            return tuple([self.__mask_sensitive_msg(arg) for arg in args])

    @overload
    def __mask_sensitive_args(self, args: None) -> None:
        pass

    @overload
    def __mask_sensitive_args(self, args: Mapping[str, object]) -> dict[str, object]:
        pass

    @overload
    def __mask_sensitive_args(self, args: Sequence[object]) -> list[object]:
        pass

    def __mask_sensitive_args(self, args: Sequence[object] | Mapping[str, object] | None) -> list[object] | dict[str, object] | None:
        """
        Masks sensitive data in the given log record arguments.

        :param args: the log record arguments, which can be a mapping, tuple, or list.
        :return: the masked log record arguments.
        """
        if isinstance(args, Mapping):
            new_args: dict[str, object] = {}
            for key in args.keys():
                # Dot-separated key names may be generated by Representor objects and should also be checked. Because
                # the key names come from python attributes, and attribute names cannot have periods in them, we don't
                # need to worry about false positives from periods with other meanings.
                if key in self.sensitive_keys or key.endswith(self.__ends_with_sensitive_keys):
                    new_args[key] = "******"
                else:
                    # mask sensitive data in dict values
                    new_args[key] = self.__mask_sensitive_msg(args[key])
            return new_args
        elif args is None:
            return args
        else:
            # when there are multiple args in record.args
            return list(self.__mask_sensitive_msg(arg) for arg in args)

    def __mask_sensitive_msg(self, message: object) -> object:
        """
        Masks sensitive data in the given log message.

        :param message: the log message, which can be a string or a dict.
        :return: the masked log message.
        """
        if not isinstance(message, str):
            if isinstance(message, (Mapping, Sequence)):
                return self.__mask_sensitive_args(message)
            if isinstance(message, URL):
                return URL(str(self.__mask_sensitive_msg(str(message))))
            return message
        for pattern_sub in self.patterns:
            if isinstance(pattern_sub, tuple):
                pattern, replace = pattern_sub
            else:
                pattern = pattern_sub
                replace = '******'
            message = pattern.sub(replace, message)
        for key in self.sensitive_keys:
            if key not in self.__sensitive_key_regexes:
                pattern_str = rf"'{key}': '[^']+'"
                replace = f"'{key}': '******'"
                self.__sensitive_key_regexes[key] = (re.compile(pattern_str), replace)
            pattern, replace = self.__sensitive_key_regexes[key]
            message = pattern.sub(replace, message)
        return message


PROD_LOG_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'sensitive_data_filter': {
            '()': SensitiveDataFilter,
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'filters': ['sensitive_data_filter'],
            'level': 'NOTSET',
            'formatter': 'default',
        },
    },
    'formatters': {
        'default': {
            'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        },
    },
    'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': True,
        },
        'botocore.endpoint': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.parsers': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.loaders': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.httpsession': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.hooks': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.utils': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.configprovider': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.regions': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.retries.standard': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.credentials': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        's3transfer.tasks': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        's3transfer.utils': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'boto3.s3.transfer': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'pymongo.command': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'pymongo.connection': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'pymongo.serverSelection': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'pymongo.topology': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'aiormq.connection': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'aio_pika.exchange': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'aio_pika.channel': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'aio_pika.robust_connection': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'aio_pika.connection': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'aio_pika.queue': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'aiosqlite': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'urllib3.connectionpool': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        }
    }
}


DEBUG_LOG_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'sensitive_data_filter': {
            '()': SensitiveDataFilter,
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'filters': ['sensitive_data_filter'],
            'level': 'NOTSET',
            'formatter': 'default',
        },
    },
    'formatters': {
        'default': {
            'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        },
    },
    'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': True,
        },
        'botocore.endpoint': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.parsers': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.loaders': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.httpsession': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.hooks': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.utils': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.configprovider': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.regions': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.retries.standard': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'botocore.credentials': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        's3transfer.tasks': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        's3transfer.utils': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'boto3.s3.transfer': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'pymongo.command': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'pymongo.connection': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'pymongo.serverSelection': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'pymongo.topology': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'aiormq.connection': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'aio_pika.exchange': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'aio_pika.channel': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'aio_pika.robust_connection': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'aio_pika.connection': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'aio_pika.queue': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'aiosqlite': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        },
        'urllib3.connectionpool': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': False,
        }
    }
}


class ScrubbingLogger(logging.Logger):
    """
    Logger that scrubs sensitive HEAObject attributes from log messages when they're passed as arguments or in a list
    or tuple that's passed as an argument.
    """

    def makeRecord(self, *args, **kwargs) -> logging.LogRecord:
        record = super().makeRecord(*args, **kwargs)
        if isinstance(record.args, tuple):
            def scrub_if_needed(obj: object) -> object:
                if isinstance(obj, HEAObject):
                    return scrubbed(obj)
                elif isinstance(obj, (list, tuple)):
                    return type(obj)(scrub_if_needed(item) for item in obj)
                else:
                    return obj
            record.args = tuple(scrub_if_needed(obj) for obj in record.args)

        return record



"""
Application properties that are loaded into the aiohttp app context when heaserver.service.runner.start is called.

HEA_REGISTRY - The base URL for the registry service.
HEA_WSTL_BUILDER_FACTORY - A zero-argument callable for getting the service's design-time WeSTL document.
HEA_DB - The database object to use.
HEA_COMPONENT - This service's base URL.
HEA_CLIENT_SESSION - A aiohttp.ClientSession instance.
HEA_MESSAGE_BROKER_PUBLISHER - The message broker publisher instance.
HEA_MESSAGE_BROKER_SUBSCRIBER - The message broker subscriber instance.
HEA_BACKGROUND_TASKS - A heaserver.service.background.BackgroundTaskManager instance.
HEA_CACHE - A cache instance for the service.
HEA_ATTRIBUTE_ENCRYPTION_KEY_GETTER - a heaserver.service.crypt.EncryptionDecryptionKeyGetter instance, currently used
    for attribute encryption and decryption, as well as configuration property decryption.
"""

HEA_REGISTRY = 'HEA_registry'
HEA_WSTL_BUILDER_FACTORY = 'HEA_WeSTL_builder_factory'
HEA_DB = 'HEA_db'
HEA_COMPONENT = 'HEA_component'
HEA_CLIENT_SESSION = 'HEA_client_session'
HEA_MESSAGE_BROKER_PUBLISHER = 'HEA_message_broker_publisher'
HEA_MESSAGE_BROKER_SUBSCRIBER = 'HEA_message_broker_subscriber'
HEA_BACKGROUND_TASKS = 'HEA_background_tasks'
HEA_CACHE = 'HEA_cache'
HEA_ATTRIBUTE_ENCRYPTION_KEY_GETTER = 'HEA_attribute_encryption_key_getter'



"""
A module providing utilities for decrypting configuration properties and for encrypting and decrypting HEAObject
attributes.
"""
from typing import Optional, cast
from heaobject.encryption import Encryption
from aiohttp import web

from .requestproperty import HEA_ATTRIBUTE_ENCRYPTION_KEY
from .appproperty import HEA_ATTRIBUTE_ENCRYPTION_KEY_GETTER
import abc


class SecretDecryption():
    """
    A helper class for decrypting encrypted configuration properties.
    """

    def __init__(self, key: bytes) -> None:
        """
        Initializes the SecretDecryption instance with the given key.

        :param key: the decryption key as bytes.
        """
        self.__encryption = Encryption(key)

    @property
    def encrypted_config_property_prefix(self) -> str:
        """
        The prefix that indicates that a configuration property is encrypted.

        :return: the encrypted property prefix.
        """
        return '{crypt}'

    def is_config_property_encrypted(self, value: str | None) -> bool:
        """
        Determines whether a configuration property is encrypted.

        :param value: the configuration property value.
        :return: True if the property is encrypted, False otherwise.
        """
        return value is not None and value.startswith(self.encrypted_config_property_prefix)

    def decrypt_config_property(self, encrypted_value: str | None) -> str | None:
        """
        Decrypts a configuration property if it is encrypted. Encrypted configuration properties are expected to be
        prefixed with '{crypt}'.

        :param encrypted_value: the encrypted configuration property value.
        :return: the decrypted configuration property value, or the original value if it was not encrypted.
        """
        prefix = self.encrypted_config_property_prefix
        if encrypted_value and encrypted_value.startswith(prefix):
            return self.__encryption.decrypt(encrypted_value[len(prefix):]).decode('utf-8')
        return encrypted_value

    def decrypt(self, data: bytes | str) -> bytes:
        """
        Decrypts the given data using the configured decryption key.

        :param data: the data to decrypt, either as bytes or a string.
        :return: the decrypted data as bytes.
        """
        return self.__encryption.decrypt(data)

    @staticmethod
    def from_request(request: web.Request) -> Optional['SecretDecryption']:
        """
        Creates a SecretDecryption instance from the request, caching the key for the duration of the request.

        :param request: the aiohttp Request.
        :return: a SecretDecryption instance, or None if the decryption key is not available.
        """
        return get_secret_decryption_from_request(request)

    @staticmethod
    def from_app(app: web.Application) -> Optional['SecretDecryption']:
        """
        Creates a SecretDecryption instance from the app.

        :param app: the aiohttp Application.
        :return: a SecretDecryption instance, or None if the decryption key is not available.
        """
        return get_secret_decryption_from_app(app)


class EncryptionDecryptionKeyGetter(abc.ABC):
    """
    An interface for getting encryption and decryption keys for attribute encryption and decryption, as well as
    configuration property decryption.
    """

    @abc.abstractmethod
    def get_attribute_encryption_key(self) -> bytes | None:
        """
        Gets the symmetric encryption key used for attribute encryption and decryption, if available.

        :return: the encryption key as bytes, or None.
        """
        pass

    @abc.abstractmethod
    def get_attribute_encryption(self) -> Encryption | None:
        """
        Gets the attribute encryption instance if an encryption key is available.

        :return: an Encryption instance, or None.
        """
        pass

    @abc.abstractmethod
    def get_secret_decryption_key(self) -> bytes | None:
        """
        Gets the decryption key for encrypted configuration properties.

        :return: the decryption key as bytes, or None.
        """
        pass

    @abc.abstractmethod
    def get_secret_decryption(self) -> SecretDecryption | None:
        """
        Gets a SecretDecryption instance for decrypting configuration properties, if a decryption key is available.

        :return: a SecretDecryption instance, or None.
        """
        pass


def get_attribute_key_from_app(app: web.Application) -> bytes | None:
    """
    Gets the symmetric key for encrypting and decrypting HEAObject attributes. Unlike the request-based version,
    this does not cache the key.

    :param app: the aiohttp Application.
    :return: the key as bytes, or None.
    """
    encryption_key_getter = cast(EncryptionDecryptionKeyGetter, app[HEA_ATTRIBUTE_ENCRYPTION_KEY_GETTER])
    return encryption_key_getter.get_secret_decryption_key()


def get_attribute_key_from_request(request: web.Request) -> bytes | None:
    """
    Gets the symmetric key for encrypting and decrypting HEAObject attributes, caching the encryption key for the
    duration of the given request.

    :param request: the aiohttp Request.
    :return: the key as bytes, or None if the key is not available.
    """
    encryption_key = cast(bytes | None, request.get(HEA_ATTRIBUTE_ENCRYPTION_KEY))
    if encryption_key is None:
        encryption_key_getter = cast(EncryptionDecryptionKeyGetter, request.app[HEA_ATTRIBUTE_ENCRYPTION_KEY_GETTER])
        encryption_key = request[HEA_ATTRIBUTE_ENCRYPTION_KEY] = encryption_key_getter.get_attribute_encryption_key()
    return encryption_key


def get_attribute_encryption_from_app(app: web.Application) -> Encryption | None:
    """
    Gets the Encryption instance for encrypting and decrypting HEAObject attributes using the encryption key getter
    stored in the app.

    :param app: the aiohttp Application.
    :return: the Encryption instance, or None if the encryption key is not available.
    """
    encryption_key = get_attribute_key_from_app(app)
    return Encryption(encryption_key) if encryption_key else None


def get_attribute_encryption_from_request(request: web.Request) -> Encryption | None:
    """
    Gets the Encryption instance for encrypting and decrypting HEAObject attributes, caching the encryption key for
    the duration of the given request.

    :param request: the aiohttp Request.
    :return: the Encryption instance, or None if the encryption key is not available.
    """
    encryption_key = get_attribute_key_from_request(request)
    return Encryption(encryption_key) if encryption_key else None


def get_secret_decryption_from_app(app: web.Application) -> SecretDecryption | None:
    """
    Gets a SecretDecryption instance using the encryption key getter stored in the app.

    :param app: the aiohttp Application.
    :return: a SecretDecryption instance, or None if the decryption key is not available.
    """
    encryption_key = get_attribute_key_from_app(app)
    return SecretDecryption(encryption_key) if encryption_key else None


def get_secret_decryption_from_request(request: web.Request) -> SecretDecryption | None:
    """
    Gets a SecretDecryption instance using the encryption key getter stored in the app, caching the encryption key
    for the duration of the request.

    :param request: the aiohttp Request.
    :return: a SecretDecryption instance, or None if the decryption key is not available.
    """
    decryption_key = get_attribute_key_from_request(request)
    return SecretDecryption(decryption_key) if decryption_key else None



"""
Starts a HEA microservice.
"""
import asyncio
import logging.config
import argparse
from aiohttp import web, TCPConnector, ClientResponseError
import os.path
from heaserver.service import appproperty, appfactory
from heaserver.service import wstl
from heaserver.service.aiohttp import client_session
from typing import Callable, Iterable, Optional, Union, AsyncIterator, Type, cast
from yarl import URL

from heaserver.service.logging import DEBUG_LOG_CONFIG, PROD_LOG_CONFIG, ScrubbingLogger

from .config import Configuration
from .db.database import MicroserviceDatabaseManager

from .defaults import DEFAULT_LOG_LEVEL, DEFAULT_PORT, DEFAULT_BASE_URL, DEFAULT_REGISTRY_URL
from ..service import client
from asyncio import create_task, wait_for, CancelledError, TimeoutError
from .backgroundtasks import BackgroundTasks
from cachetools import TTLCache
from enum import Enum

routes = web.RouteTableDef()


def init(port: Union[int, str] = DEFAULT_PORT,
         base_url: Union[URL, str] = DEFAULT_BASE_URL,
         config_file: Optional[str] = None,
         config_string: Optional[str] = None,
         logging_level: Optional[str] = None) -> Configuration:
    """
    Sets a logging level and python's "basic" logging config, and it returns a new Configuration object. To customize
    logging further, use the heaserver.service.logging module to set up logging before calling this function and
    omit the logging_level argument. Alternatively, instead of calling this function, use the
    heaserver.service.runner.init_cmd_line() function to configure logging from command line arguments.

    :param port: the port the service will listen to. If omitted, the DEFAULT_PORT is used.
    :param base_url: the service's base URL. If omitted, the DEFAULT_BASE_URL is used.
    :param config_file:
    :param config_string:
    :param logging_level: configure the logging level. If omitted, logging will not be configured.
    :return: a Configuration object.
    """
    if logging_level is not None:
        logging.basicConfig(level=logging_level)
    config = Configuration(base_url=base_url,
                           port=port,
                           config_file=config_file,
                           config_str=config_string)
    return config


def init_cmd_line(description: str = 'A HEA microservice', default_port: int = DEFAULT_PORT) -> Configuration:
    """
    Parses command line arguments and configures the service accordingly. Must be called before anything else, if you
    want to use command line arguments to configure the service. It exits the program if any command line arguments are
    invalid. It configures logging according to the command line arguments, or uses the default logging configuration
    for the specified environment, or uses the default logging configuration if no logging configuration file or
    environment is specified.

    :param description: optional description that will appear when the script is passed the -h or --help option.
    :param default_port: the optional port on which to listen if none is specified by command line argument. If omitted
    and no port is specified by command line argument, port 8080 is used.
    :return: a Configuration object.
    :raises ValueError: if any arguments are invalid.

    The following command line arguments are accepted:
        -b or --baseurl, which optionally sets the base URL of the service. If unspecified, the default is
        http://localhost:<port>, where port is the 8080 or the port number set by the --port argument.
        -p or --port, which optionally sets the port the microservice will listen on. The default is 8080.
        -f or --configuration, which optionally sets an INI file to use for additional configuration.
        -l or --logging, which optionally sets a standard logging configuration INI file. If unspecified, the
            DEFAULT_LOG_LEVEL variable will be used to set the default log level.
        -e or --env, which optionally sets the runtime environment (development, staging, or production). Parsing is
            case-insensitive. When unspecified, the default value is development. It currently impacts logging
            configuration, and only if no logging configuration file is specified.

    Microservices must not call logging.getLogger until init has been called, or logging will not be configured
    properly.

    The INI configuration file is parsed by the built-in configparser module and may contain the following settings:

    ;Base URL for the HEA registry service (default is http://localhost:8080/heaserver-server-registry)
    Registry = url of the HEA registry service

    ;See the documentation for the db object that you passed in for the config file properties that it expects.
    """
    assert default_port is not None

    parser = argparse.ArgumentParser(description)
    parser.add_argument('-b', '--baseurl',
                        metavar='baseurl',
                        type=str,
                        default=str(DEFAULT_BASE_URL),
                        help='The base URL of the service')
    parser.add_argument('-p', '--port',
                        metavar='port',
                        type=int,
                        default=default_port,
                        help='The port on which the server will listen for connections')
    parser.add_argument('-f', '--configuration',
                        metavar='configuration',
                        type=str,
                        help='Path to a HEA configuration file in INI format')
    parser.add_argument('-l', '--logging',
                        metavar='logging',
                        type=str,
                        help='Standard logging configuration file')
    parser.add_argument('-e', '--env',
                        metavar='env',
                        type=str, choices=['development', 'staging', 'production'],
                        help='The runtime environment (development, staging, or production). Parsing is case-insensitive. When unspecified, the default value is development. It currently impacts logging configuration, and only if no logging configuration file is specified.')
    args = parser.parse_args()

    _configure_logging_from_args(args)

    baseurl_ = args.baseurl
    return Configuration(base_url=baseurl_[:-1] if baseurl_.endswith('/') else baseurl_,
                         config_file=args.configuration,
                         port=args.port)


def start(package_name: str | None = None,
          db: Optional[Type[MicroserviceDatabaseManager]] = None,
          wstl_builder_factory: Optional[Callable] = None,
          cleanup_ctx: Optional[Iterable[Callable[[web.Application], AsyncIterator[None]]]] = None,
          config: Optional[Configuration] = None) -> None:
    """
    Starts the microservice. It calls get_application() to get the AioHTTP app
    object, sets up a global HTTP client session object in the appproperty.HEA_CLIENT_SESSION property, and then calls
    AioHTTP's aiohttp.web.run_app() method with it to launch the service. It sets the
    application and request properties described in the documentation for get_application(). It sets uvloop as the
    event loop on platforms that uvloop supports. The service listens on all available network interfaces (IP address
    0.0.0.0), and the port is set by passing a heaserver.service.config.Configuration object with the desired port
    attribute. If the port is unset, the service listens on port 8080. Limiting the network interfaces that the service
    listens on is not supported but can be achieved with firewall rules, Docker, or other network configuration tools.

    :param package_name: the microservice's distribution package name. The HEA server framework uses this argument to
    register metadata about the microservice in the HEA registry service. Omit this argument if you do not want this
    microservice to register metadata about itself using this mechanism. For example, the registry service has to
    register itself using its own mechanism.
    :param db: a database manager type from the heaserver.server.db package, if database connectivity is needed. Sets
    the appproperty.HEA_DB application property to a database object created by this database manager.
    :param wstl_builder_factory: a zero-argument callable that will return a design-time WeSTL document. Optional if
    this service has no actions.
    :param cleanup_ctx: an iterable of asynchronous iterators that will be passed into the aiohttp cleanup context.
    The iterator should have a single yield statement that separates code to be run upon startup and code to be
    run upon shutdown. The shutdown code will run only if the startup code did not raise any exceptions. The startup
    code will run in sequential order. The shutdown code will run in reverse order.
    :param config: a Configuration instance.
    :raises ValueError: if registering the externally facing URL with the registry service failed.

    This function must be called after init.

    A 'registry' property is set in the application context with the registry service's base URL.
    """
    _logger = logging.getLogger(__name__)
    if config is None:
        config_ = init()
    else:
        config_ = config
    db_ = db() if db is not None else None

    app = get_application(package_name, db=db_, wstl_builder_factory=wstl_builder_factory,
                          cleanup_ctx=cleanup_ctx, config=config_)
    app.cleanup_ctx.append(client_session_cleanup_ctx)
    try:
        import uvloop  # type:ignore[import-not-found]
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        _logger.debug('Running uvloop event loop')
    except:
        _logger.debug('Running default event loop')
    web.run_app(app, port=config_.port if config_.port else DEFAULT_PORT)


def get_application(package_name: str | None = None,
                    db: Optional[MicroserviceDatabaseManager] = None,
                    wstl_builder_factory: Optional[Callable[[], wstl.RuntimeWeSTLDocumentBuilder]] = None,
                    cleanup_ctx: Optional[Iterable[Callable[[web.Application], AsyncIterator[None]]]] = None,
                    config: Optional[Configuration] = None) -> web.Application:
    """
    Gets the aiohttp application object for this microservice. It is called by start() during normal operations, and
    by test cases when running tests.

    It registers cleanup context iterators that set the following application properties:
    HEA_DB: a database object from the heaserver.server.db package.
    HEA_CLIENT_SESSION: a aiohttp.web.ClientSession HTTP client object. This property is only set if the testing
    argument is False. If running test cases, testing should be set to True, and the HEAAioHTTPTestCase class will
    handle creating and destroying HTTP clients instead.
    HEA_REGISTRY: The base URL for the registry service.
    HEA_COMPONENT: This service's base URL.
    HEA_WSTL_BUILDER_FACTORY: the wstl_builder_factory argument.

    :param package_name: the microservice's distribution package name. If None, the service will not register its
    externally facing base URL with the registry service. This may be what you want in testing scenarios in which no
    registry service is running.
    :param db: a database object from the heaserver.server.db package, if database connectivity is needed.
    :param wstl_builder_factory: a zero-argument callable that will return a RuntimeWeSTLDocumentBuilder. Optional if
    this service has no actions. Typically, you will use the heaserver.service.wstl.get_builder_factory function to
    get a factory object.
    :param cleanup_ctx: an iterable of asynchronous iterators that will be passed into the aiohttp cleanup context.
    The iterators should have a single yield statement that separates code to be run upon startup and code to be
    run upon shutdown. The shutdown code will run only if the startup code did not raise any exceptions. Cleanup
    context iterators cannot assume that any of the above application properties are available.
    :param config: a Configuration instance.
    :return: an aiohttp application object.
    :raises ValueError: if registering the externally facing URL with the registry service failed.

    This function must be called after init, and it is called by start.

    A 'registry' property is set in the application context with the registry service's base URL.
    """
    logger = logging.getLogger(__name__)

    logger.info('Starting HEA')

    app = appfactory.new_app()

    if hasattr(asyncio, 'eager_task_factory'):
        async def _set_eager_task_factory(app: web.Application) -> AsyncIterator[None]:
            loop = asyncio.get_running_loop()
            try:
                loop.set_task_factory(asyncio.eager_task_factory)   # type: ignore[attr-defined]
                logger.debug('Eager task factory is turned on')
            except AttributeError:
                pass
            yield
        app.cleanup_ctx.append(_set_eager_task_factory)
    else:
        logger.debug('Using default task factory because eager task factory is not supported by this version of python')

    if db:
        async def _db(app: web.Application) -> AsyncIterator[None]:
            if db:
                async with db.database() as database:
                    app[appproperty.HEA_DB] = database
                    yield
            else:
                app[appproperty.HEA_DB] = None
        if config:
            db.config = config
        app.cleanup_ctx.append(_db)

    def gen_hea_registry_app_property():
        registry_url = config.parsed_config['DEFAULT'].get('Registry', DEFAULT_REGISTRY_URL) if config else None
        app[appproperty.HEA_REGISTRY] = registry_url
    gen_hea_registry_app_property()

    # Instance of heaserver.service.crypt.EncryptionDecryptionKeyGetter. We currently use only one key for attribute
    # encryption and decryption, as well as configuration property decryption.
    app[appproperty.HEA_ATTRIBUTE_ENCRYPTION_KEY_GETTER] = config

    async def _hea_component(app: web.Application) -> AsyncIterator[None]:
        external_base_url = str(config.base_url) if config else None
        app[appproperty.HEA_COMPONENT] = external_base_url
        if package_name is not None:
            # Needs its own client session because the main client session has not started yet.
            logger.info('Registering external URL %s for component %s with HEA registry service at %s...',
                        external_base_url, package_name, app[appproperty.HEA_REGISTRY])
            async with client_session(connector=TCPConnector(), connector_owner=True, raise_for_status=True) as session:
                logger.debug('Created temporary client session to register component %s with HEA registry service',
                             package_name)
                # Consults the registry service as the system|none user, so the heaserver-people service does not need
                # to be consulted for permissions information.
                component = await client.get_component_by_name(app, package_name, client_session=session)
                if component is not None:
                    logger.debug('Found component %r in registry service to register external URL', component)
                    component.external_base_url = external_base_url
                    logger.info('Registering external URL %s for component %s...', external_base_url, package_name,)
                    try:
                        assert component.id is not None, 'component has a None id'
                        await client.put(app, URL(app[appproperty.HEA_REGISTRY]) / 'components' / component.id, component, client_session=session)
                        logger.info('Successfully registered external URL for component %s.', package_name)
                    except ClientResponseError as e:
                        raise ValueError(f'Failed to register external URL for component {package_name}') from e
                else:
                    logger.warning('The external URL for component %s is not registered with the HEA registry service.', package_name)
        else:
            logger.info('Skipping registration of the external URL for component %s with the HEA registry service.', external_base_url)
        yield

    app.cleanup_ctx.append(_hea_component)

    async def _background_tasks(app: web.Application) -> AsyncIterator[None]:
        logger.info('Starting background tasks...')
        background_tasks = BackgroundTasks(app)
        app[appproperty.HEA_BACKGROUND_TASKS] = background_tasks
        async def background_tasks_coro():
            await background_tasks.auto_join()
        task = create_task(background_tasks_coro(), name='background_tasks')
        logger.info('Background tasks started.')
        yield
        logger.info('Shutting down background tasks...')
        task.cancel()
        try:
            await wait_for(task, timeout=30)
        except CancelledError:
            pass
        except TimeoutError:
            logger.warning('Timed out trying to end background tasks: %s', background_tasks)
        finally:
            await background_tasks.clear()
        logger.info('Background tasks shut down.')

    app.cleanup_ctx.append(_background_tasks)

    async def _cache(app: web.Application) -> AsyncIterator[None]:
        cache: TTLCache = TTLCache(maxsize=128, ttl=30)
        app[appproperty.HEA_CACHE] = cache
        yield
        app[appproperty.HEA_CACHE] = None


    app.cleanup_ctx.append(_cache)

    app.add_routes(routes)

    async def _hea_wstl_builder_factory(app: web.Application) -> AsyncIterator[None]:
        if wstl_builder_factory is None:
            logger.debug('No design-time WeSTL loader was provided.')
            wstl_builder_factory_ = wstl.builder_factory()
        else:
            wstl_builder_factory_ = wstl_builder_factory

        app[appproperty.HEA_WSTL_BUILDER_FACTORY] = wstl_builder_factory_
        yield

    app.cleanup_ctx.append(_hea_wstl_builder_factory)

    if cleanup_ctx is not None:
        for cb in cleanup_ctx:
            app.cleanup_ctx.append(cb)
    return app


async def client_session_cleanup_ctx(app: web.Application) -> AsyncIterator[None]:
    """
    Manages the global HTTP client session, to be added to the cleanup context.

    :param app: the AioHTTP Application object.
    :return: an async iterator.
    """
    _logger = logging.getLogger(__name__)
    _logger.debug('Starting client session')
    async with client_session(connector=TCPConnector(), connector_owner=True, raise_for_status=True) as session:
        app[appproperty.HEA_CLIENT_SESSION] = session
        _logger.debug('Client session started')
        yield
        _logger.debug('Closing client session')


def scheduled_cleanup_ctx(coro, delay: int) -> Callable[[web.Application], AsyncIterator[None]]:
    """
    A cleanup context factory for scheduling recurring background tasks with a delay

    :param coro: The coroutine to be scheduled
    :param delay: The delay in seconds before scheduling the coroutine. The delay and scheduling will repeat
    indefinitely.
    :return: a callable that can be added to the cleanup context.
    """
    logger = logging.getLogger(__name__)
    async def scheduler(app: web.Application) -> AsyncIterator[None]:
        async def loop():
            while True:
                await asyncio.sleep(delay)
                if (tasks := cast(BackgroundTasks, app.get(appproperty.HEA_BACKGROUND_TASKS))) is not None:
                    await tasks.add(coro)
                else:
                    logger.warning('No background tasks manager found; skipping scheduled task execution. The '
                                   'background tasks manager may just not have been initialized yet. If this keeps '
                                   'happening, however, there may be a problem.')

        task = asyncio.create_task(loop())
        yield
        task.cancel()
        await task
    return scheduler

def wait_on_coro_cleanup_ctx(coro, delay: int = 0):
    """
    A producer to schedules reoccurring background tasks to be awaited before adding another

    :param coro: The coroutine to be scheduled
    :param delay: The delay in seconds to run the reoccurring coroutine
    """
    logger = logging.getLogger(__name__)
    async def scheduler(app: web.Application):

        async def loop():
            while True:
                await asyncio.sleep(delay)
                if (tasks := cast(BackgroundTasks, app.get(appproperty.HEA_BACKGROUND_TASKS))) is not None:
                    task_name = await tasks.add(coro)
                    await tasks.join(task_name)
                    if tasks.failed(task_name):
                        logger.debug('One background tasks failed. continuing...')
                else:
                    logger.warning('No background tasks manager found; skipping scheduled task execution. The '
                                   'background tasks manager may just not have been initialized yet. If this keeps '
                                   'happening, however, there may be a problem.')

        task = asyncio.create_task(loop())
        yield
        task.cancel()
        await task
    return scheduler

scheduled_cleanup_ctx_factory = scheduled_cleanup_ctx


def _configure_logging_from_args(args: argparse.Namespace) -> None:
    """
    Configures logging according to command line arguments. It looks for args.logging and args.env to determine how
    to configure logging. If neither is specified, it uses the default logging configuration. If both are specified,
    args.logging takes precedence and reads from the specified logging configuration file. If only args.env is
    specified, it uses the logging configuration from the heaserver.service.logging module for the specified
    environment.

    :param args: the argparse.Namespace object returned by parsing command line arguments.
    """
    class LoggingConfigType(Enum):
        """Types of logging configuration sources."""
        FILE = 10
        ENV = 20
        DEFAULT = 30

    logging.setLoggerClass(ScrubbingLogger)
    logging_config_type = LoggingConfigType.DEFAULT
    if args.logging and os.path.isfile(args.logging):
        logging_config_type = LoggingConfigType.FILE
        logging.config.fileConfig(args.logging, disable_existing_loggers=False)
    elif args.env:
        logging_config_type = LoggingConfigType.ENV
        env = args.env.lower()
        if env == 'development':
            logging.config.dictConfig(DEBUG_LOG_CONFIG)
        elif env in ('staging', 'production'):
            logging.config.dictConfig(PROD_LOG_CONFIG)
        else:
            raise RuntimeError(f'Unreachable condition: unknown environment {args.env}')
    else:
        logging.config.dictConfig(DEBUG_LOG_CONFIG)

    logger = logging.getLogger(__name__)

    match logging_config_type:
        case LoggingConfigType.FILE:
            logger.info('Using logging configuration file %s', args.logging)
        case LoggingConfigType.ENV:
            logger.info('Using %s logging configuration', args.env)
        case LoggingConfigType.DEFAULT:
            logger.info('Using default logging configuration')

    if args.configuration:
        logger.info('Parsing HEA config file %s', args.configuration)
    else:
        logger.info('No HEA config file found')

    if args.logging and os.path.isfile(args.logging):
        logger.info('Parsing logging config file %s', args.logging)
    else:
        logger.info('No logging config file found')



import unittest

import heaserver.service.defaults
from heaserver.service import runner
from unittest.mock import patch
from unittest.result import TestResult as _TestResult  # Keeps pytest from trying to parse it as a test case.
from typing import Optional
from heaserver.service.config import Configuration
import sys
from pathlib import Path
from tempfile import NamedTemporaryFile
from cryptography.fernet import Fernet


class TestDefaultConfiguration(unittest.TestCase):

    def run(self, result: Optional[_TestResult] = None) -> Optional[_TestResult]:
        """
        Patches sys.argv with an empty list before running each test. This keeps pytest command line argument parsing
        from conflicting with HEA command line argument parsing.

        :param result: a TestResult object for compiling which tests passed and failed.
        :return: the same TestResult object, with the results of another test recorded.
        """
        with patch.object(sys, 'argv', []):
            return super().run(result)

    def setUp(self) -> None:
        self.config = runner.init_cmd_line(description='foo')

    def test_default_base_url(self) -> None:
        self.assertEqual(heaserver.service.defaults.DEFAULT_BASE_URL, self.config.base_url)


class TestCustomConfiguration(unittest.TestCase):

    def run(self, result: Optional[_TestResult] = None) -> Optional[_TestResult]:
        """
        Patches sys.argv with an empty list before running each test. This keeps pytest command line argument parsing
        from conflicting with HEA command line argument parsing.

        :param result: a TestResult object for compiling which tests passed and failed.
        :return: the same TestResult object, with the results of another test recorded.
        """
        with patch.object(sys, 'argv', []):
            return super().run(result)

    def setUp(self) -> None:
        self.config_str = """
[DEFAULT]
Registry = http://customhost:9090/hea-service
EncryptionKey = Xx0TnPIeA4B2jokub4HMjKkIIqDPBdEpPFtD8gD70Fg=
"""
        self.config = Configuration(config_str=self.config_str)

    def test_custom_registry(self) -> None:
        self.assertEqual('http://customhost:9090/hea-service', self.config.registry_url)

    def test_encryption_key(self) -> None:
        self.assertEqual(b'Xx0TnPIeA4B2jokub4HMjKkIIqDPBdEpPFtD8gD70Fg=', self.config.get_attribute_encryption_key())


class TestCustomConfigurationWithEncryptionKeyFile(unittest.TestCase):

    def run(self, result: Optional[_TestResult] = None) -> Optional[_TestResult]:
        """
        Patches sys.argv with an empty list before running each test. This keeps pytest command line argument parsing
        from conflicting with HEA command line argument parsing.

        :param result: a TestResult object for compiling which tests passed and failed.
        :return: the same TestResult object, with the results of another test recorded.
        """
        with patch.object(sys, 'argv', []):
            return super().run(result)

    def setUp(self) -> None:
        with NamedTemporaryFile(delete=False) as temp_file:
            self.temp_file = temp_file
            self.key = Fernet.generate_key()
            self.temp_file.write(self.key)

        self.config_str = f"""
[DEFAULT]
Registry = http://customhost:9090/hea-service
EncryptionKeyFile = {self.temp_file.name}
"""
        self.config = Configuration(config_str=self.config_str)

    def tearDown(self):
        Path(self.temp_file.name).unlink(missing_ok=True)

    def test_encryption_key_file(self) -> None:
        self.assertEqual(self.temp_file.name, self.config.encryption_key_file)

    def test_encryption_key_from_file(self) -> None:
        self.assertEqual(self.key, self.config.get_attribute_encryption_key())
